# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logitech_receiver import settings as _settings
from logitech_receiver import settings_templates as _settings_templates
from solaar import configuration as _configuration


def _print_setting(s, verbose=True):
    print('#', s.label)
    if verbose:
        if s.description:
            print('#', s.description.replace('\n', ' '))
        if s.kind == _settings.KIND.toggle:
            print('#   possible values: on/true/t/yes/y/1 or off/false/f/no/n/0')
        elif s.kind == _settings.KIND.choice:
            print(
                '#   possible values: one of [', ', '.join(str(v) for v in s.choices),
                '], or higher/lower/highest/max/lowest/min'
            )
        else:
            # wtf?
            pass
    value = s.read(cached=False)
    if value is None:
        print(s.name, '= ? (failed to read from device)')
    else:
        print(s.name, '= %s' % value)


def to_int(s):
    try:
        return int(s)
    except ValueError:
        return None


def select_choice(value, choices, setting, key):
    lvalue = value.lower()
    ivalue = to_int(value)
    val = None
    for choice in choices:
        if value == str(choice):
            val = choice
            break
    if val is not None:
        value = val
    elif ivalue is not None and ivalue >= 1 and ivalue <= len(choices):
        value = choices[ivalue - 1]
    elif lvalue in ('higher', 'lower'):
        old_value = setting.read() if key is None else setting.read_key(key)
        if old_value is None:
            raise Exception("%s: could not read current value'" % setting.name)
        if lvalue == 'lower':
            lower_values = choices[:old_value]
            value = lower_values[-1] if lower_values else choices[:][0]
        elif lvalue == 'higher':
            higher_values = choices[old_value + 1:]
            value = higher_values[0] if higher_values else choices[:][-1]
    elif lvalue in ('highest', 'max', 'first'):
        value = choices[:][-1]
    elif lvalue in ('lowest', 'min', 'last'):
        value = choices[:][0]
    else:
        raise Exception('%s: possible values are [%s]' % (setting.name, ', '.join(str(v) for v in choices)))
    return value


def select_toggle(value, setting):
    try:
        value = bool(int(value))
    except Exception:
        if value.lower() in ('true', 'yes', 'on', 't', 'y'):
            value = True
        elif value.lower() in ('false', 'no', 'off', 'f', 'n'):
            value = False
        else:
            raise Exception("%s: don't know how to interpret '%s' as boolean" % (setting.name, value))
    return value


def select_range(value, setting):
    try:
        value = int(value)
    except ValueError:
        raise Exception("%s: can't interpret '%s' as integer" % (setting.name, value))
    min, max = setting.range
    if value < min or value > max:
        raise Exception("%s: value '%s' out of bounds" % (setting.name, value))
    return value


def run(receivers, args, find_receiver, find_device):
    assert receivers
    assert args.device

    device_name = args.device.lower()

    dev = None
    for dev in find_device(receivers, device_name):
        if dev.ping():
            break
        dev = None

    if not dev:
        raise Exception("no online device found matching '%s'" % device_name)

    if not args.setting:  # print all settings, so first set them all up
        if not dev.settings:
            raise Exception('no settings for %s' % dev.name)
        _configuration.attach_to(dev)
        _settings.apply_all_settings(dev)
        print(dev.name, '(%s) [%s:%s]' % (dev.codename, dev.wpid, dev.serial))
        for s in dev.settings:
            print('')
            _print_setting(s)
        return

    setting_name = args.setting.lower()
    setting = _settings_templates.check_feature_setting(dev, setting_name)
    if setting is None:
        raise Exception("no setting '%s' for %s" % (args.setting, dev.name))
    _configuration.attach_to(dev)

    if args.value_key is None:
        setting.apply()
        _print_setting(setting)
        return

    if setting.kind == _settings.KIND.toggle:
        value = select_toggle(args.value_key, setting)
        print('Setting %s of %s to %s' % (setting.name, dev.name, value))
        result = setting.write(value)

    elif setting.kind == _settings.KIND.range:
        value = select_range(args.value_key, setting)
        print('Setting %s of %s to %s' % (setting.name, dev.name, value))
        result = setting.write(value)

    elif setting.kind == _settings.KIND.choice:
        value = select_choice(args.value_key, setting.choices, setting, None)
        print('Setting %s of %s to %s' % (setting.name, dev.name, value))
        result = setting.write(value)

    elif setting.kind == _settings.KIND.map_choice:
        key = args.value_key
        ikey = to_int(key)
        k = next((k for k in setting.choices.keys() if key == k), None)
        if k is None and ikey is not None:
            k = next((k for k in setting.choices.keys() if ikey == k), None)
        if k is not None:
            value = select_choice(args.extra_subkey, setting.choices[k], setting, key)
        else:
            raise Exception("%s: key '%s' not in setting" % (setting.name, key))
        print('Setting %s of %s key %r to %r' % (setting.name, dev.name, k, value))
        result = setting.write_key_value(int(k), value)

    elif setting.kind == _settings.KIND.multiple_toggle:
        key = args.value_key
        ikey = to_int(key)
        k = next((k for k in setting._labels if key == k), None)
        if k is None and ikey is not None:
            k = next((k for k in setting._labels if ikey == k), None)
        if k is not None:
            value = select_toggle(args.extra_subkey, setting)
        else:
            raise Exception("%s: key '%s' not in setting" % (setting.name, key))
        print('Setting %s key %r to %r' % (setting.name, k, value))
        result = setting.write_key_value(int(k), value)

    elif setting.kind == _settings.KIND.multiple_range:
        key = args.value_key
        ikey = to_int(key)
        if args.extra_subkey is None:
            raise Exception('%s: setting needs a subkey' % (setting.name))
        if args.extra2 is None or to_int(args.extra2) is None:
            raise Exception('%s: setting needs an integer value, not %s' % (setting.name, args.extra2))
        if not setting._value:  # ensure that there are values to look through
            setting.read()
        k = next((k for k in setting._value if key == k), None)
        if k is None and ikey is not None:
            k = next((k for k in setting._value if ikey == k), None)
        item = setting._value[k]
        if args.extra_subkey in item.keys():
            item[args.extra_subkey] = to_int(args.extra2)
        else:
            raise Exception("%s: key '%s' not in setting" % (setting.name, key))
        print('Setting %s key %s parameter %s to %r' % (setting.name, k, args.extra_subkey, item[args.extra_subkey]))
        result = setting.write_item_value(int(k), item)


# KIND = _NamedInts(, multiple_toggle=0x10, multiple_range=0x40)
# BitField; MultipleRange
# disable_keyboard_keys, gesture2; gesture2_params
# k400+ / craft, k400+; k400+

    else:
        raise Exception('NotImplemented')

    if result is None:
        raise Exception("%s: failed to set value '%s' [%r]" % (setting.name, str(value), value))



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os as _os
import os.path as _path

from json import dump as _json_save
from json import load as _json_load
from logging import DEBUG as _DEBUG
from logging import INFO as _INFO
from logging import getLogger

from solaar import __version__

_log = getLogger(__name__)
del getLogger

_XDG_CONFIG_HOME = _os.environ.get('XDG_CONFIG_HOME') or _path.expanduser(_path.join('~', '.config'))
_file_path = _path.join(_XDG_CONFIG_HOME, 'solaar', 'config.json')

_KEY_VERSION = '_version'
_KEY_NAME = '_name'
_KEY_SERIAL = '_serial'
_KEY_MODEL_ID = '_modelId'
_KEY_UNIT_ID = '_unitId'
_configuration = {}


def _load():
    if _path.isfile(_file_path):
        loaded_configuration = {}
        try:
            with open(_file_path, 'r') as config_file:
                loaded_configuration = _json_load(config_file)
        except Exception:
            _log.error('failed to load from %s', _file_path)

        # loaded_configuration.update(_configuration)
        _configuration.clear()
        _configuration.update(loaded_configuration)

    if _log.isEnabledFor(_DEBUG):
        _log.debug('load => %s', _configuration)

    _cleanup(_configuration)
    _cleanup_load(_configuration)
    _configuration[_KEY_VERSION] = __version__
    return _configuration


def save():
    # don't save if the configuration hasn't been loaded
    if _KEY_VERSION not in _configuration:
        return

    dirname = _os.path.dirname(_file_path)
    if not _path.isdir(dirname):
        try:
            _os.makedirs(dirname)
        except Exception:
            _log.error('failed to create %s', dirname)
            return False

    _cleanup(_configuration)

    try:
        with open(_file_path, 'w') as config_file:
            _json_save(_configuration, config_file, skipkeys=True, indent=2, sort_keys=True)

        if _log.isEnabledFor(_INFO):
            _log.info('saved %s to %s', _configuration, _file_path)
        return True
    except Exception as e:
        _log.error('failed to save to %s: %s', _file_path, e)


def _cleanup(d):
    # remove None values from the dict
    for key in list(d.keys()):
        value = d.get(key)
        if value is None:
            del d[key]
        elif isinstance(value, dict):
            _cleanup(value)


def _cleanup_load(d):
    # remove boolean values for mouse-gesture and dpi-sliding
    for device in d.values():
        if isinstance(device, dict):
            for setting in ['mouse-gestures', 'dpi-sliding']:
                mg = device.get(setting, None)
                if mg is True or mg is False:
                    del device[setting]


class _DeviceEntry(dict):
    def __init__(self, device, **kwargs):
        super(_DeviceEntry, self).__init__(**kwargs)
        if self.get(_KEY_NAME) != device.name:
            self[_KEY_NAME] = device.name
        self.update(device)

    def __setitem__(self, key, value):
        super(_DeviceEntry, self).__setitem__(key, value)
        save()

    def update(self, device):
        if device.modelId and device.modelId != self.get(_KEY_MODEL_ID):
            self[_KEY_MODEL_ID] = device.modelId
        if device.unitId and device.unitId != self.get(_KEY_UNIT_ID):
            self[_KEY_UNIT_ID] = device.unitId
        if device.serial and device.serial != '?' and device.serial != self.get(_KEY_SERIAL):
            self[_KEY_SERIAL] = device.serial

    def get_sensitivity(self, name):
        return self.get('_sensitive', {}).get(name, False)

    def set_sensitivity(self, name, value):
        sensitives = self.get('_sensitive', {})
        if sensitives.get(name) != value:
            sensitives[name] = value
            self.__setitem__('_sensitive', sensitives)


# This is neccessarily complicate because the same device can be attached in several different ways.
# All HID++ 2.0 devices have a modelId and unitId, which can be accessed when they are connected.
# When paired via a receiver the receiver provides a WPID and a serial number.
# The unitId and serial number are supposed to be the same, but for some models they are not
# so even though the modelId includes the WPID it is not always possible to determine the identity of a
# paired but not receiver-connected device for which the unitId is not known.
# This only happens is Solaar has never seen the device while it is paired and connected through a receiver.
def persister(device):
    if not _configuration:
        _load()

    entry = {}
    key = None
    if device.wpid:  # connected via receiver
        entry = _configuration.get('%s:%s' % (device.wpid, device.serial), {})
    if entry or device.protocol == 1.0:  # found entry or create entry for old-style devices
        key = '%s:%s' % (device.wpid, device.serial)
    elif not entry and device.modelId:  # online new-style device so look for modelId and unitId
        for k, c in _configuration.items():
            if isinstance(c, dict) and c.get(_KEY_MODEL_ID) == device.modelId and c.get(_KEY_UNIT_ID) == device.unitId:
                entry = c  # use the entry that matches modelId and unitId
                key = k
                break
        if device.wpid and entry:  # move entry to wpid:serial
            del _configuration[key]
            key = '%s:%s' % (device.wpid, device.serial)
            _configuration[key] = entry
        elif device.wpid and not entry:  # create now with wpid:serial
            key = '%s:%s' % (device.wpid, device.serial)
        elif not entry:  # create now with modelId:unitId
            key = '%s:%s' % (device.modelId, device.unitId)
    else:  # defer until more is known (i.e., device comes on line)
        return

    if key and not isinstance(entry, _DeviceEntry):
        entry = _DeviceEntry(device, **entry)
        _configuration[key] = entry
    if isinstance(entry, _DeviceEntry):
        entry.update(device)

    return entry


def attach_to(device):
    pass



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from threading import Timer as _Timer

from gi.repository import Gdk, GLib, Gtk
from logitech_receiver.settings import KIND as _SETTING_KIND
from logitech_receiver.settings import SENSITIVITY_IGNORE as _SENSITIVITY_IGNORE
from solaar.i18n import _, ngettext
from solaar.ui import ui_async as _ui_async

#
#
#


def _read_async(setting, force_read, sbox, device_is_online, sensitive):
    def _do_read(s, force, sb, online, sensitive):
        v = s.read(not force)
        GLib.idle_add(_update_setting_item, sb, v, online, sensitive, priority=99)

    _ui_async(_do_read, setting, force_read, sbox, device_is_online, sensitive)


def _write_async(setting, value, sbox):
    failed, spinner, control = _get_failed_spinner_control(sbox)
    control.set_sensitive(False)
    failed.set_visible(False)
    spinner.set_visible(True)
    spinner.start()

    def _do_write(s, v, sb):
        v = setting.write(v)
        GLib.idle_add(_update_setting_item, sb, v, True, priority=99)

    _ui_async(_do_write, setting, value, sbox)


def _write_async_key_value(setting, key, value, sbox):
    failed, spinner, control = _get_failed_spinner_control(sbox)
    control.set_sensitive(False)
    failed.set_visible(False)
    spinner.set_visible(True)
    spinner.start()

    def _do_write_key_value(s, k, v, sb):
        v = setting.write_key_value(k, v)
        GLib.idle_add(_update_setting_item, sb, {k: v}, True, priority=99)

    _ui_async(_do_write_key_value, setting, key, value, sbox)


def _write_async_item_value(setting, item, value, sbox):
    failed, spinner, control = _get_failed_spinner_control(sbox)
    control.set_sensitive(False)
    failed.set_visible(False)
    spinner.set_visible(True)
    spinner.start()

    def _do_write_item_value(s, k, v, sb):
        v = setting.write_item_value(k, v)
        GLib.idle_add(_update_setting_item, sb, {k: v}, True, priority=99)

    _ui_async(_do_write_item_value, setting, item, value, sbox)


#
#
#


def _create_toggle_control(setting):
    def _switch_notify(switch, _ignore, s):
        if switch.get_sensitive():
            _write_async(s, switch.get_active() is True, switch.get_parent())

    c = Gtk.Switch()
    c.connect('notify::active', _switch_notify, setting)
    return c


def _create_choice_control(setting):
    def _combo_notify(cbbox, s):
        if cbbox.get_sensitive():
            _write_async(s, cbbox.get_active_id(), cbbox.get_parent())

    c = Gtk.ComboBoxText()
    # TODO i18n text entries
    for entry in setting.choices:
        c.append(str(int(entry)), str(entry))
    c.connect('changed', _combo_notify, setting)
    return c


def _create_map_choice_control(setting):
    def _map_value_notify_key(cbbox, s):
        setting, valueBox = s
        key_choice = int(cbbox.get_active_id())
        if cbbox.get_sensitive():
            valueBox.remove_all()
            _map_populate_value_box(valueBox, setting, key_choice)

    def _map_value_notify_value(cbbox, s):
        setting, keyBox = s
        key_choice = keyBox.get_active_id()
        if key_choice is not None and cbbox.get_sensitive() and cbbox.get_active_id():
            if setting._value.get(key_choice) != int(cbbox.get_active_id()):
                setting._value[key_choice] = int(cbbox.get_active_id())
                _write_async_key_value(setting, key_choice, setting._value[key_choice], cbbox.get_parent().get_parent())

    def _map_populate_value_box(valueBox, setting, key_choice):
        choices = None
        choices = setting.choices[key_choice]
        current = setting._value.get(str(key_choice)) if setting._value else None
        if choices:
            # TODO i18n text entries
            for choice in choices:
                valueBox.append(str(int(choice)), str(choice))
            if current is not None:
                valueBox.set_active_id(str(int(current)))

    c = Gtk.HBox(homogeneous=False, spacing=6)
    keyBox = Gtk.ComboBoxText()
    valueBox = Gtk.ComboBoxText()
    c.pack_start(keyBox, False, False, 0)
    c.pack_end(valueBox, False, False, 0)
    # TODO i18n text entries
    for entry in setting.choices:
        keyBox.append(str(int(entry)), str(entry))
    keyBox.set_active(0)
    keyBox.connect('changed', _map_value_notify_key, (setting, valueBox))
    _map_populate_value_box(valueBox, setting, int(keyBox.get_active_id()))
    valueBox.connect('changed', _map_value_notify_value, (setting, keyBox))
    return c


def _create_slider_control(setting):
    class SliderControl:
        __slots__ = ('gtk_range', 'timer', 'setting')

        def __init__(self, setting):
            self.setting = setting
            self.timer = None

            self.gtk_range = Gtk.Scale()
            self.gtk_range.set_range(*self.setting.range)
            self.gtk_range.set_round_digits(0)
            self.gtk_range.set_digits(0)
            self.gtk_range.set_increments(1, 5)
            self.gtk_range.connect('value-changed', lambda _, c: c._changed(), self)

        def _write(self):
            _write_async(self.setting, int(self.gtk_range.get_value()), self.gtk_range.get_parent())
            self.timer.cancel()

        def _changed(self):
            if self.gtk_range.get_sensitive():
                if self.timer:
                    self.timer.cancel()
                self.timer = _Timer(0.5, lambda: GLib.idle_add(self._write))
                self.timer.start()

    control = SliderControl(setting)
    return control.gtk_range


def _create_multiple_toggle_control(setting, change):
    def _toggle_notify(control, _, setting):
        if control.get_sensitive():
            key = control._setting_key
            new_state = control.get_active()
            if setting._value[key] != new_state:
                setting._value[key] = new_state
                p = control
                for _ in range(5):
                    p = p.get_parent()
                _write_async_key_value(setting, key, new_state, p)

    def _toggle_display(lb):
        lb._showing = not lb._showing
        if not lb._showing:
            for c in lb.get_children():
                c.hide()
            lb.hide()
        else:
            lb.show()
            for c in lb.get_children():
                c.show_all()

    lb = Gtk.ListBox()
    lb._toggle_display = (lambda l: (lambda: _toggle_display(l)))(lb)
    lb._showing = True
    lb.set_selection_mode(Gtk.SelectionMode.NONE)
    lb.set_no_show_all(True)
    lb._label_control_pairs = []
    btn = Gtk.Button('? / ?')
    for k in setting._validator.all_options():
        h = Gtk.HBox(homogeneous=False, spacing=0)
        lbl_text = str(k)
        lbl_tooltip = None
        if hasattr(setting, '_labels'):
            l1, l2 = setting._labels.get(k, (None, None))
            if l1:
                lbl_text = l1
            if l2:
                lbl_tooltip = l2
        lbl = Gtk.Label(lbl_text)
        h.set_tooltip_text(lbl_tooltip or ' ')
        control = Gtk.Switch()
        control._setting_key = str(int(k))
        control.connect('notify::active', _toggle_notify, setting)
        h.pack_start(lbl, False, False, 0)
        h.pack_end(control, False, False, 0)
        lbl.set_alignment(0.0, 0.5)
        lbl.set_margin_left(30)
        lb.add(h)
        lb._label_control_pairs.append((lbl, control))
    _disable_listbox_highlight_bg(lb)
    lb._toggle_display()
    btn.connect('clicked', lambda _: lb._toggle_display())

    hbox = Gtk.HBox(homogeneous=False, spacing=6)
    hbox.pack_end(change, False, False, 0)
    hbox.pack_end(btn, False, False, 0)
    btn.set_alignment(1.0, 0.5)
    vbox = Gtk.VBox(homogeneous=False, spacing=6)
    vbox.pack_start(hbox, True, True, 0)
    vbox.pack_end(lb, True, True, 0)
    vbox._header, vbox._button, vbox._control = hbox, btn, lb
    return vbox


def _create_multiple_range_control(setting, change):
    def _write(control, setting, item, sub_item):
        control._timer.cancel()
        delattr(control, '_timer')
        new_state = int(control.get_value())
        if setting._value[str(int(item))][str(sub_item)] != new_state:
            setting._value[str(int(item))][str(sub_item)] = new_state
            p = control
            for _i in range(7):
                p = p.get_parent()
            _write_async_item_value(setting, str(int(item)), setting._value[str(int(item))], p)

    def _changed(control, setting, item, sub_item):
        if control.get_sensitive():
            if hasattr(control, '_timer'):
                control._timer.cancel()
            control._timer = _Timer(0.5, lambda: GLib.idle_add(_write, control, setting, item, sub_item))
            control._timer.start()

    def _toggle_display(lb):
        lb._showing = not lb._showing
        if not lb._showing:
            for c in lb.get_children():
                c.hide()
            lb.hide()
        else:
            lb.show()
            for c in lb.get_children():
                c.show_all()

    lb = Gtk.ListBox()
    lb._toggle_display = (lambda l: (lambda: _toggle_display(l)))(lb)
    lb.set_selection_mode(Gtk.SelectionMode.NONE)
    lb._showing = True
    lb.set_no_show_all(True)
    lb._items = []
    btn = Gtk.Button('...')
    for item in setting._validator.items:
        lbl_text = str(item)
        lbl_tooltip = None
        if hasattr(setting, '_labels'):
            l1, l2 = setting._labels.get(item, (None, None))
            if l1:
                lbl_text = l1
            if l2:
                lbl_tooltip = l2
        item_lbl = Gtk.Label(lbl_text)
        lb.add(item_lbl)
        lb.set_tooltip_text(lbl_tooltip or ' ')
        item_lb = Gtk.ListBox()
        item_lb.set_selection_mode(Gtk.SelectionMode.NONE)
        item_lb._sub_items = []
        for sub_item in setting._validator.sub_items[item]:
            h = Gtk.HBox(homogeneous=False, spacing=20)
            lbl_text = str(sub_item)
            lbl_tooltip = None
            if hasattr(setting, '_labels_sub'):
                l1, l2 = setting._labels_sub.get(str(sub_item), (None, None))
                if l1:
                    lbl_text = l1
                if l2:
                    lbl_tooltip = l2
            sub_item_lbl = Gtk.Label(lbl_text)
            h.set_tooltip_text(lbl_tooltip or ' ')
            h.pack_start(sub_item_lbl, False, False, 0)
            sub_item_lbl.set_margin_left(30)
            sub_item_lbl.set_alignment(0.0, 0.5)
            if sub_item.widget == 'Scale':
                control = Gtk.Scale.new_with_range(Gtk.Orientation.HORIZONTAL, sub_item.minimum, sub_item.maximum, 1)
                control.set_round_digits(0)
                control.set_digits(0)
                h.pack_end(control, True, True, 0)
            elif sub_item.widget == 'SpinButton':
                control = Gtk.SpinButton.new_with_range(sub_item.minimum, sub_item.maximum, 1)
                control.set_digits(0)
                h.pack_end(control, False, False, 0)
            else:
                raise NotImplementedError
            control.connect('value-changed', _changed, setting, item, sub_item)
            item_lb.add(h)
            h._setting_sub_item = sub_item
            h._label, h._control = sub_item_lbl, control
            item_lb._sub_items.append(h)
        item_lb._setting_item = item
        _disable_listbox_highlight_bg(item_lb)
        lb.add(item_lb)
        lb._items.append(item_lb)
    _disable_listbox_highlight_bg(lb)
    lb._toggle_display()
    btn.connect('clicked', lambda _: lb._toggle_display())
    btn.set_alignment(1.0, 0.5)
    hbox = Gtk.HBox(homogeneous=False, spacing=6)
    hbox.pack_end(change, False, False, 0)
    hbox.pack_end(btn, False, False, 0)
    vbox = Gtk.VBox(homogeneous=False, spacing=6)
    vbox.pack_start(hbox, True, True, 0)
    vbox.pack_end(lb, True, True, 0)
    vbox._header, vbox._button, vbox._control = hbox, btn, lb
    return vbox


#
#
#

_allowables_icons = {True: 'changes-allow', False: 'changes-prevent', _SENSITIVITY_IGNORE: 'dialog-error'}
_allowables_tooltips = {
    True: _('Changes allowed'),
    False: _('No changes allowed'),
    _SENSITIVITY_IGNORE: _('Ignore this setting')
}
_next_allowable = {True: False, False: _SENSITIVITY_IGNORE, _SENSITIVITY_IGNORE: True}
_icons_allowables = {v: k for k, v in _allowables_icons.items()}


# clicking on the lock icon changes from changeable to unchangeable to ignore
def _change_click(eb, button, arg):
    control, device, name = arg
    icon = eb.get_children()[0]
    icon_name, _ = icon.get_icon_name()
    allowed = _icons_allowables.get(icon_name, True)
    new_allowed = _next_allowable[allowed]
    control.set_sensitive(new_allowed is True)
    _change_icon(new_allowed, icon)
    if device.persister:  # remember the new setting sensitivity
        device.persister.set_sensitivity(name, new_allowed)
    if allowed == _SENSITIVITY_IGNORE:  # update setting if it was being ignored
        setting = next((s for s in device.settings if s.name == name), None)
        if setting:
            persisted = device.persister.get(setting.name) if device.persister else None
            if persisted is not None:
                _write_async(setting, persisted, control.get_parent())
            else:
                _read_async(setting, True, control.get_parent(), bool(device.online), control.get_sensitive())
    return True


def _change_icon(allowed, icon):
    allowed = allowed if allowed in _allowables_icons else True
    icon.set_from_icon_name(_allowables_icons[allowed], Gtk.IconSize.LARGE_TOOLBAR)
    icon.set_tooltip_text(_allowables_tooltips[allowed])


def _create_sbox(s, device):
    sbox = Gtk.HBox(homogeneous=False, spacing=6)
    lbl = Gtk.Label(s.label)
    label = Gtk.EventBox()
    label.add(lbl)

    spinner = Gtk.Spinner()
    spinner.set_tooltip_text(_('Working') + '...')

    failed = Gtk.Image.new_from_icon_name('dialog-warning', Gtk.IconSize.SMALL_TOOLBAR)
    failed.set_tooltip_text(_('Read/write operation failed.'))

    change_icon = Gtk.Image.new_from_icon_name('changes-prevent', Gtk.IconSize.LARGE_TOOLBAR)
    _change_icon(False, change_icon)
    change = Gtk.EventBox()
    change.add(change_icon)

    if s.kind == _SETTING_KIND.toggle:
        control = _create_toggle_control(s)
    elif s.kind == _SETTING_KIND.choice:
        control = _create_choice_control(s)
    elif s.kind == _SETTING_KIND.range:
        control = _create_slider_control(s)
    elif s.kind == _SETTING_KIND.map_choice:
        control = _create_map_choice_control(s)
    elif s.kind == _SETTING_KIND.multiple_toggle:
        vbox = _create_multiple_toggle_control(s, change)
        control = vbox._control
        lbl.set_alignment(0.0, 0.5)
        sbox.pack_start(vbox, True, True, 0)
    elif s.kind == _SETTING_KIND.multiple_range:
        vbox = _create_multiple_range_control(s, change)
        control = vbox._control
        lbl.set_alignment(0.0, 0.5)
        sbox.pack_start(vbox, True, True, 0)
    else:
        raise Exception('NotImplemented')
    control.set_sensitive(False)  # the first read will enable it
    control.kind = s.kind

    change.set_sensitive(True)
    change.connect('button-press-event', _change_click, (control, device, s.name))

    if s.kind in [_SETTING_KIND.multiple_toggle, _SETTING_KIND.multiple_range]:
        vbox._header.pack_start(label, False, False, 0)
        vbox._header.pack_end(spinner, False, False, 0)
        vbox._header.pack_end(failed, False, False, 0)
        sbox._button = vbox._button
    else:
        sbox.pack_start(label, False, False, 0)
        sbox.pack_end(change, False, False, 0)
        sbox.pack_end(control, s.kind == _SETTING_KIND.range, s.kind == _SETTING_KIND.range, 0)
        sbox.pack_end(spinner, False, False, 0)
        sbox.pack_end(failed, False, False, 0)
    sbox._label = label
    sbox._lbl = lbl
    sbox._spinner = spinner
    sbox._failed = failed
    sbox._change = change
    sbox._change_icon = change_icon
    sbox._control = control

    if s.description:
        sbox.set_tooltip_text(s.description)

    sbox.show_all()

    spinner.start()  # the first read will stop it
    failed.set_visible(False)

    return sbox


def _update_setting_item(sbox, value, is_online=True, sensitive=True):
    failed, spinner, control = _get_failed_spinner_control(sbox)
    spinner.set_visible(False)
    spinner.stop()

    if value is None:
        control.set_sensitive(False)
        _change_icon(False, sbox._change_icon)
        failed.set_visible(is_online)
        return

    control.set_sensitive(False)
    failed.set_visible(False)
    if isinstance(control, Gtk.Switch):
        control.set_active(value)
    elif isinstance(control, Gtk.ComboBoxText):
        control.set_active_id(str(int(value)))
    elif isinstance(control, Gtk.Scale):
        control.set_value(int(value))
    elif isinstance(control, Gtk.HBox):
        kbox, vbox = control.get_children()  # depends on box layout
        if value.get(kbox.get_active_id()) is not None:
            vbox.set_active_id(str(value.get(kbox.get_active_id())))
    elif isinstance(control, Gtk.ListBox):
        if control.kind == _SETTING_KIND.multiple_toggle:
            total = len(control._label_control_pairs)
            active = 0
            to_join = []
            for lbl, elem in control._label_control_pairs:
                v = value.get(elem._setting_key, None)
                if v is not None:
                    elem.set_active(v)
                if elem.get_active():
                    active += 1
                to_join.append(lbl.get_text() + ': ' + str(elem.get_active()))
            b = ', '.join(to_join)
            sbox._button.set_label(f'{active} / {total}')
            sbox._button.set_tooltip_text(b)
        elif control.kind == _SETTING_KIND.multiple_range:
            b = ''
            n = 0
            for ch in control._items:
                # item
                item = ch._setting_item
                v = value.get(str(int(item)), None)
                if v is not None:
                    b += str(item) + ': ('
                    to_join = []
                    for c in ch._sub_items:
                        # sub-item
                        sub_item = c._setting_sub_item
                        c._control.set_value(v[str(sub_item)])
                        n += 1
                        to_join.append(str(sub_item) + f'={v[str(sub_item)]}')
                    b += ', '.join(to_join) + ') '
                lbl_text = ngettext('%d value', '%d values', n) % n
                sbox._button.set_label(lbl_text)
                sbox._button.set_tooltip_text(b)
        else:
            raise NotImplementedError
    else:
        raise Exception('NotImplemented')

    control.set_sensitive(sensitive is True)
    _change_icon(sensitive, sbox._change_icon)


def _get_failed_spinner_control(sbox):
    return sbox._failed, sbox._spinner, sbox._control


def _disable_listbox_highlight_bg(lb):
    colour = Gdk.RGBA()
    colour.parse('rgba(0,0,0,0)')
    for child in lb.get_children():
        child.override_background_color(Gtk.StateFlags.PRELIGHT, colour)


#
#
#

# config panel
_box = None
_items = {}


def create():
    global _box
    assert _box is None
    _box = Gtk.VBox(homogeneous=False, spacing=8)
    _box._last_device = None
    return _box


def update(device, is_online=None):
    assert _box is not None
    assert device
    device_id = (device.receiver.path if device.receiver else device.path, device.number)
    if is_online is None:
        is_online = bool(device.online)

    # if the device changed since last update, clear the box first
    if device_id != _box._last_device:
        _box.set_visible(False)
        _box._last_device = device_id

    # hide controls belonging to other devices
    for k, sbox in _items.items():
        sbox = _items[k]
        sbox.set_visible(k[0:2] == device_id)

    for s in device.settings:
        k = (device_id[0], device_id[1], s.name)
        if k in _items:
            sbox = _items[k]
        else:
            sbox = _items[k] = _create_sbox(s, device)
            _box.pack_start(sbox, False, False, 0)

        sensitive = device.persister.get_sensitivity(s.name) if device.persister else True
        _read_async(s, False, sbox, is_online, sensitive)

    _box.set_visible(True)


def clean(device):
    """Remove the controls for a given device serial.
    Needed after the device has been unpaired.
    """
    assert _box is not None
    device_id = (device.receiver.path if device.receiver else device.path, device.number)
    for k in list(_items.keys()):
        if k[0:2] == device_id:
            _box.remove(_items[k])
            del _items[k]


def destroy():
    global _box
    _box = None
    _items.clear()



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

import math

from copy import copy as _copy
from logging import DEBUG as _DEBUG
from logging import getLogger

from . import hidpp20 as _hidpp20
from . import special_keys as _special_keys
from .common import NamedInt as _NamedInt
from .common import NamedInts as _NamedInts
from .common import bytes2int as _bytes2int
from .common import int2bytes as _int2bytes
from .common import unpack as _unpack

_log = getLogger(__name__)
del getLogger

#
#
#

SENSITIVITY_IGNORE = 'ignore'
KIND = _NamedInts(toggle=0x01, choice=0x02, range=0x04, map_choice=0x0A, multiple_toggle=0x10, multiple_range=0x40)


class Setting(object):
    """A setting descriptor.
    Needs to be instantiated for each specific device."""
    __slots__ = (
        'name', 'label', 'description', 'kind', 'device_kind', 'feature', 'persist', '_rw', '_validator', '_callback',
        '_device', '_value'
    )

    def __init__(self, name, rw, validator=None, callback=None, kind=None, device_kind=None, feature=None, persist=True):
        assert name
        self.name = name[0]
        self.label = name[1]
        self.description = name[2]
        self.device_kind = device_kind
        self.feature = getattr(rw, 'feature', None)
        self.persist = persist
        self._rw = rw
        assert (validator and not callback) or (not validator and callback)
        self._validator = validator
        self._callback = callback

        assert kind is None or validator is None or kind & validator.kind != 0
        self.kind = kind or getattr(validator, 'kind', None)

    def __call__(self, device):
        assert not hasattr(self, '_value')
        # combined keyboards and touchpads (e.g., K400) break this assertion so don't use it
        # assert self.device_kind is None or device.kind in self.device_kind
        p = device.protocol
        if p == 1.0:
            # HID++ 1.0 devices do not support features
            assert self._rw.kind == RegisterRW.kind
        elif p >= 2.0:
            # HID++ 2.0 devices do not support registers
            assert self._rw.kind == FeatureRW.kind
        o = _copy(self)
        if o._callback:
            o._validator = o._callback(device)
            if o._validator is None:
                return None
            assert o.kind is None or o.kind & o._validator.kind != 0
            o.kind = o.kind or o._validator.kind
        o._value = None
        o._device = device
        return o

    @property
    def choices(self):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')

        return self._validator.choices if self._validator and self._validator.kind & KIND.choice else None

    @property
    def range(self):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')

        if self._validator.kind == KIND.range:
            return (self._validator.min_value, self._validator.max_value)

    def _pre_read(self, cached, key=None):
        if self.persist and self._value is None and getattr(self._device, 'persister', None):
            # We haven't read a value from the device yet,
            # maybe we have something in the configuration.
            self._value = self._device.persister.get(self.name)
        if cached and self._value is not None:
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                # If this is a new device (or a new setting for an old device),
                # make sure to save its current value for the next time.
                self._device.persister[self.name] = self._value

    def read(self, cached=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings read %r from %s', self.name, self._value, self._device)

        self._pre_read(cached)

        if cached and self._value is not None:
            return self._value

        if self._device.online:
            reply = self._rw.read(self._device)
            if reply:
                self._value = self._validator.validate_read(reply)
            if self.persist and self._device.persister and self.name not in self._device.persister:
                # Don't update the persister if it already has a value,
                # otherwise the first read might overwrite the value we wanted.
                self._device.persister[self.name] = self._value
            return self._value

    def _pre_write(self, save=True):
        # Remember the value we're trying to set, even if the write fails.
        # This way even if the device is offline or some other error occurs,
        # the last value we've tried to write is remembered in the configuration.
        if self.persist and self._device.persister and save:
            self._device.persister[self.name] = self._value

    def write(self, value, save=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert value is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings write %r to %s', self.name, value, self._device)

        if self._device.online:
            if self._value != value:
                self._value = value
                self._pre_write(save)

            current_value = None
            if self._validator.needs_current_value:
                # the validator needs the current value, possibly to merge flag values
                current_value = self._rw.read(self._device)

            data_bytes = self._validator.prepare_write(value, current_value)
            if data_bytes is not None:
                if _log.isEnabledFor(_DEBUG):
                    _log.debug('%s: settings prepare write(%s) => %r', self.name, value, data_bytes)

                reply = self._rw.write(self._device, data_bytes)
                if not reply:
                    # tell whomever is calling that the write failed
                    return None

            return value

    def apply(self):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: apply %s (%s)', self.name, self._value, self._device)

        value = self.read(self.persist)  # Don't use persisted value if setting doesn't persist
        if self.persist and value is not None:  # If setting doesn't persist no need to write value just read
            self.write(value, save=False)

    def __str__(self):
        if hasattr(self, '_value'):
            assert hasattr(self, '_device')
            return '<Setting([%s:%s] %s:%s=%s)>' % (
                self._rw.kind, self._validator.kind if self._validator else None, self._device.codename, self.name, self._value
            )
        return '<Setting([%s:%s] %s)>' % (self._rw.kind, self._validator.kind if self._validator else None, self.name)

    __unicode__ = __repr__ = __str__


class Settings(Setting):
    """A setting descriptor for multiple choices, being a map from keys to values.
    Needs to be instantiated for each specific device."""
    def read(self, cached=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings read %r from %s', self.name, self._value, self._device)

        self._pre_read(cached)

        if cached and self._value is not None:
            return self._value

        if self._device.online:
            reply_map = {}
            for key in self._validator.choices:
                reply = self._rw.read(self._device, key)
                if reply:
                    # keys are ints, because that is what the device uses,
                    # encoded into strings because JSON requires strings as keys
                    reply_map[str(int(key))] = self._validator.validate_read(reply, key)
            self._value = reply_map
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                # Don't update the persister if it already has a value,
                # otherwise the first read might overwrite the value we wanted.
                self._device.persister[self.name] = self._value
            return self._value

    def read_key(self, key, cached=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert key is not None
        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings read %r key %r from %s', self.name, self._value, key, self._device)

        self._pre_read(cached)
        if cached and self._value is not None:
            return self._value[str(int(key))]

        if self._device.online:
            reply = self._rw.read(self._device, key)
            if reply:
                self._value[str(int(key))] = self._validator.validate_read(reply, key)
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                self._device.persister[self.name] = self._value
            return self._value[str(int(key))]

    def write(self, map, save=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert map is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings write %r to %s', self.name, map, self._device)

        if self._device.online:
            self._value = map
            self._pre_write(save)
            for key, value in map.items():
                data_bytes = self._validator.prepare_write(int(key), value)
                if data_bytes is not None:
                    if _log.isEnabledFor(_DEBUG):
                        _log.debug('%s: settings prepare map write(%s,%s) => %r', self.name, key, value, data_bytes)
                    reply = self._rw.write(self._device, int(key), data_bytes)
                    if not reply:
                        return None
            return map

    def write_key_value(self, key, value):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert key is not None
        assert value is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings write key %r value %r to %s', self.name, key, value, self._device)

        if self._device.online:
            if not self._value:
                self.read()
            try:
                data_bytes = self._validator.prepare_write(int(key), value)
                # always need to write to configuration because dictionary is shared and could have changed
                self._value[str(key)] = value
                self._pre_write()
            except ValueError:
                data_bytes = value = None
            if data_bytes is not None:
                if _log.isEnabledFor(_DEBUG):
                    _log.debug('%s: settings prepare key value write(%s,%s) => %r', self.name, key, value, data_bytes)
                reply = self._rw.write(self._device, int(key), data_bytes)
                if not reply:
                    # tell whomever is calling that the write failed
                    return None
            return value


class LongSettings(Setting):
    """A setting descriptor for multiple choices, being a map from keys to values.
    Allows multiple write requests, if the options don't fit in 16 bytes.
    The validator must return a list.
    Needs to be instantiated for each specific device."""
    def read(self, cached=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings read %r from %s', self.name, self._value, self._device)

        self._pre_read(cached)

        if cached and self._value is not None:
            return self._value

        if self._device.online:
            reply_map = {}
            # Reading one item at a time. This can probably be optimised
            for item in self._validator.items:
                r = self._validator.prepare_read_item(item)
                reply = self._rw.read(self._device, r)
                if reply:
                    # keys are ints, because that is what the device uses,
                    # encoded into strings because JSON requires strings as keys
                    reply_map[str(int(item))] = self._validator.validate_read_item(reply, item)
            self._value = reply_map
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                # Don't update the persister if it already has a value,
                # otherwise the first read might overwrite the value we wanted.
                self._device.persister[self.name] = self._value
            return self._value

    def read_item(self, item, cached=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert item is not None
        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings read %r item %r from %s', self.name, self._value, item, self._device)

        self._pre_read(cached)
        if cached and self._value is not None:
            return self._value[str(int(item))]

        if self._device.online:
            r = self._validator.prepare_read_item(item)
            reply = self._rw.read(self._device, r)
            if reply:
                self._value[str(int(item))] = self._validator.validate_read_item(reply, item)
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                self._device.persister[self.name] = self._value
            return self._value[str(int(item))]

    def write(self, map, save=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert map is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings write %r to %s', self.name, map, self._device)
        if self._device.online:
            self._value = map
            self._pre_write(save)
            for item, value in map.items():
                data_bytes_list = self._validator.prepare_write(self._value)
                if data_bytes_list is not None:
                    for data_bytes in data_bytes_list:
                        if data_bytes is not None:
                            if _log.isEnabledFor(_DEBUG):
                                _log.debug('%s: settings prepare map write(%s,%s) => %r', self.name, item, value, data_bytes)
                            reply = self._rw.write(self._device, data_bytes)
                            if not reply:
                                return None
            return map

    def write_item_value(self, item, value):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert item is not None
        assert value is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings write item %r value %r to %s', self.name, item, value, self._device)

        if self._device.online:
            if not self._value:
                self.read()
            data_bytes = self._validator.prepare_write_item(item, value)
            self._value[str(int(item))] = value
            self._pre_write()
            if data_bytes is not None:
                if _log.isEnabledFor(_DEBUG):
                    _log.debug('%s: settings prepare item value write(%s,%s) => %r', self.name, item, value, data_bytes)
                reply = self._rw.write(self._device, data_bytes)
                if not reply:
                    return None
            return value


class BitFieldSetting(Setting):
    """A setting descriptor for a set of choices represented by one bit each, being a map from options to booleans.
    Needs to be instantiated for each specific device."""
    def read(self, cached=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings read %r from %s', self.name, self._value, self._device)

        self._pre_read(cached)

        if cached and self._value is not None:
            return self._value

        if self._device.online:
            reply_map = {}
            reply = self._do_read()
            if reply:
                # keys are ints, because that is what the device uses,
                # encoded into strings because JSON requires strings as keys
                reply_map = self._validator.validate_read(reply)
            self._value = reply_map
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                # Don't update the persister if it already has a value,
                # otherwise the first read might overwrite the value we wanted.
                self._device.persister[self.name] = self._value
            return self._value

    def _do_read(self):
        return self._rw.read(self._device)

    def read_key(self, key, cached=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert key is not None
        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings read %r key %r from %s', self.name, self._value, key, self._device)

        self._pre_read(cached)

        if cached and self._value is not None:
            return self._value[str(int(key))]

        if self._device.online:
            reply = self._do_read_key(key)
            if reply:
                self._value = self._validator.validate_read(reply)
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                self._device.persister[self.name] = self._value
            return self._value[str(int(key))]

    def _do_read_key(self, key):
        return self._rw.read(self._device, key)

    def write(self, map, save=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert map is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings write %r to %s', self.name, map, self._device)
        if self._device.online:
            self._value = map
            self._pre_write(save)
            data_bytes = self._validator.prepare_write(self._value)
            if data_bytes is not None:
                if _log.isEnabledFor(_DEBUG):
                    _log.debug('%s: settings prepare map write(%s) => %r', self.name, self._value, data_bytes)
                # if prepare_write returns a list, write one item at a time
                seq = data_bytes if isinstance(data_bytes, list) else [data_bytes]
                for b in seq:
                    reply = self._rw.write(self._device, b)
                    if not reply:
                        return None
            return map

    def write_key_value(self, key, value):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert key is not None
        assert value is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings write key %r value %r to %s', self.name, key, value, self._device)

        if self._device.online:
            if not self._value:
                self.read()
            value = bool(value)
            self._value[str(key)] = value
            self._pre_write()

            data_bytes = self._validator.prepare_write(self._value)
            if data_bytes is not None:
                if _log.isEnabledFor(_DEBUG):
                    _log.debug('%s: settings prepare key value write(%s,%s) => %r', self.name, key, str(value), data_bytes)
                # if prepare_write returns a list, write one item at a time
                seq = data_bytes if isinstance(data_bytes, list) else [data_bytes]
                for b in seq:
                    reply = self._rw.write(self._device, b)
                    if not reply:
                        return None

            return value


class BitFieldWithOffsetAndMaskSetting(BitFieldSetting):
    """A setting descriptor for a set of choices represented by one bit each,
    each one having an offset, being a map from options to booleans.
    Needs to be instantiated for each specific device."""
    def _do_read(self):
        return {r: self._rw.read(self._device, r) for r in self._validator.prepare_read()}

    def _do_read_key(self, key):
        r = self._validator.prepare_read_key(key)
        return {r: self._rw.read(self._device, r)}


#
# read/write low-level operators
#


class RegisterRW(object):
    __slots__ = ('register', )

    kind = _NamedInt(0x01, 'register')

    def __init__(self, register):
        assert isinstance(register, int)
        self.register = register

    def read(self, device):
        return device.read_register(self.register)

    def write(self, device, data_bytes):
        return device.write_register(self.register, data_bytes)


class FeatureRW(object):
    __slots__ = ('feature', 'read_fnid', 'write_fnid', 'no_reply')

    kind = _NamedInt(0x02, 'feature')
    default_read_fnid = 0x00
    default_write_fnid = 0x10

    def __init__(self, feature, read_fnid=default_read_fnid, write_fnid=default_write_fnid, no_reply=False):
        assert isinstance(feature, _NamedInt)
        self.feature = feature
        self.read_fnid = read_fnid
        self.write_fnid = write_fnid
        self.no_reply = no_reply

    def read(self, device, data_bytes=b''):
        assert self.feature is not None
        return device.feature_request(self.feature, self.read_fnid, data_bytes)

    def write(self, device, data_bytes):
        assert self.feature is not None
        reply = device.feature_request(self.feature, self.write_fnid, data_bytes, no_reply=self.no_reply)
        return reply if not self.no_reply else True


class FeatureRWMap(FeatureRW):
    kind = _NamedInt(0x02, 'feature')
    default_read_fnid = 0x00
    default_write_fnid = 0x10
    default_key_byte_count = 1

    def __init__(
        self,
        feature,
        read_fnid=default_read_fnid,
        write_fnid=default_write_fnid,
        key_byte_count=default_key_byte_count,
        no_reply=False
    ):
        assert isinstance(feature, _NamedInt)
        self.feature = feature
        self.read_fnid = read_fnid
        self.write_fnid = write_fnid
        self.key_byte_count = key_byte_count
        self.no_reply = no_reply

    def read(self, device, key):
        assert self.feature is not None
        key_bytes = _int2bytes(key, self.key_byte_count)
        return device.feature_request(self.feature, self.read_fnid, key_bytes)

    def write(self, device, key, data_bytes):
        assert self.feature is not None
        key_bytes = _int2bytes(key, self.key_byte_count)
        reply = device.feature_request(self.feature, self.write_fnid, key_bytes, data_bytes, no_reply=self.no_reply)
        return reply if not self.no_reply else True


#
# value validators
# handle the conversion from read bytes, to setting value, and back
#


class BooleanValidator(object):
    __slots__ = ('true_value', 'false_value', 'mask', 'needs_current_value')

    kind = KIND.toggle
    default_true = 0x01
    default_false = 0x00
    # mask specifies all the affected bits in the value
    default_mask = 0xFF

    def __init__(self, true_value=default_true, false_value=default_false, mask=default_mask):
        if isinstance(true_value, int):
            assert isinstance(false_value, int)
            if mask is None:
                mask = self.default_mask
            else:
                assert isinstance(mask, int)
            assert true_value & false_value == 0
            assert true_value & mask == true_value
            assert false_value & mask == false_value
            self.needs_current_value = (mask != self.default_mask)
        elif isinstance(true_value, bytes):
            if false_value is None or false_value == self.default_false:
                false_value = b'\x00' * len(true_value)
            else:
                assert isinstance(false_value, bytes)
            if mask is None or mask == self.default_mask:
                mask = b'\xFF' * len(true_value)
            else:
                assert isinstance(mask, bytes)
            assert len(mask) == len(true_value) == len(false_value)
            tv = _bytes2int(true_value)
            fv = _bytes2int(false_value)
            mv = _bytes2int(mask)
            assert tv != fv  # true and false might be something other than bit values
            assert tv & mv == tv
            assert fv & mv == fv
            self.needs_current_value = any(m != b'\xFF' for m in mask)
        else:
            raise Exception("invalid mask '%r', type %s" % (mask, type(mask)))

        self.true_value = true_value
        self.false_value = false_value
        self.mask = mask

    def validate_read(self, reply_bytes):
        if isinstance(self.mask, int):
            reply_value = ord(reply_bytes[:1]) & self.mask
            if _log.isEnabledFor(_DEBUG):
                _log.debug('BooleanValidator: validate read %r => %02X', reply_bytes, reply_value)
            if reply_value == self.true_value:
                return True
            if reply_value == self.false_value:
                return False
            _log.warn(
                'BooleanValidator: reply %02X mismatched %02X/%02X/%02X', reply_value, self.true_value, self.false_value,
                self.mask
            )
            return False

        count = len(self.mask)
        mask = _bytes2int(self.mask)
        reply_value = _bytes2int(reply_bytes[:count]) & mask

        true_value = _bytes2int(self.true_value)
        if reply_value == true_value:
            return True

        false_value = _bytes2int(self.false_value)
        if reply_value == false_value:
            return False

        _log.warn('BooleanValidator: reply %r mismatched %r/%r/%r', reply_bytes, self.true_value, self.false_value, self.mask)
        return False

    def prepare_write(self, new_value, current_value=None):
        if new_value is None:
            new_value = False
        else:
            assert isinstance(new_value, bool)

        to_write = self.true_value if new_value else self.false_value

        if isinstance(self.mask, int):
            if current_value is not None and self.needs_current_value:
                to_write |= ord(current_value[:1]) & (0xFF ^ self.mask)
            if current_value is not None and to_write == ord(current_value[:1]):
                return None
        else:
            to_write = bytearray(to_write)
            count = len(self.mask)
            for i in range(0, count):
                b = ord(to_write[i:i + 1])
                m = ord(self.mask[i:i + 1])
                assert b & m == b
                # b &= m
                if current_value is not None and self.needs_current_value:
                    b |= ord(current_value[i:i + 1]) & (0xFF ^ m)
                to_write[i] = b
            to_write = bytes(to_write)

            if current_value is not None and to_write == current_value[:len(to_write)]:
                return None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('BooleanValidator: prepare_write(%s, %s) => %r', new_value, current_value, to_write)

        return to_write


class BitFieldValidator(object):
    __slots__ = ('byte_count', 'options')

    kind = KIND.multiple_toggle

    def __init__(self, options, byte_count=None):
        assert (isinstance(options, list))
        self.options = options
        self.byte_count = (max(x.bit_length() for x in options) + 7) // 8
        if byte_count:
            assert (isinstance(byte_count, int) and byte_count >= self.byte_count)
            self.byte_count = byte_count

    def validate_read(self, reply_bytes):
        r = _bytes2int(reply_bytes[:self.byte_count])
        value = {str(int(k)): False for k in self.options}
        m = 1
        for _ in range(8 * self.byte_count):
            if m in self.options:
                value[str(int(m))] = bool(r & m)
            m <<= 1
        return value

    def prepare_write(self, new_value):
        assert (isinstance(new_value, dict))
        w = 0
        for k, v in new_value.items():
            if v:
                w |= int(k)
        return _int2bytes(w, self.byte_count)

    def all_options(self):
        return self.options


class BitFieldWithOffsetAndMaskValidator(object):
    __slots__ = ('byte_count', 'options', '_option_from_key', '_mask_from_offset', '_option_from_offset_mask')

    kind = KIND.multiple_toggle
    sep = 0x01

    def __init__(self, options, byte_count=None):
        assert (isinstance(options, list))
        # each element of options must have .offset and .mask,
        # and its int representation must be its id (not its index)
        self.options = options
        # to retrieve the options efficiently:
        self._option_from_key = {}
        self._mask_from_offset = {}
        self._option_from_offset_mask = {}
        for opt in options:
            self._option_from_key[int(opt)] = opt
            try:
                self._mask_from_offset[opt.offset] |= opt.mask
            except KeyError:
                self._mask_from_offset[opt.offset] = opt.mask
            try:
                mask_to_opt = self._option_from_offset_mask[opt.offset]
            except KeyError:
                mask_to_opt = {}
                self._option_from_offset_mask[opt.offset] = mask_to_opt
            mask_to_opt[opt.mask] = opt
        self.byte_count = (max(x.mask.bit_length() for x in options) + 7) // 8
        if byte_count:
            assert (isinstance(byte_count, int) and byte_count >= self.byte_count)
            self.byte_count = byte_count

    def prepare_read(self):
        r = []
        for offset, mask in self._mask_from_offset.items():
            b = (offset << (8 * (self.byte_count + 1)))
            b |= (self.sep << (8 * self.byte_count)) | mask
            r.append(_int2bytes(b, self.byte_count + 2))
        return r

    def prepare_read_key(self, key):
        option = self._option_from_key.get(key, None)
        if option is None:
            return None
        b = option.offset << (8 * (self.byte_count + 1))
        b |= (self.sep << (8 * self.byte_count)) | option.mask
        return _int2bytes(b, self.byte_count + 2)

    def validate_read(self, reply_bytes_dict):
        values = {str(int(k)): False for k in self.options}
        for query, b in reply_bytes_dict.items():
            offset = _bytes2int(query[0:1])
            b += (self.byte_count - len(b)) * b'\x00'
            value = _bytes2int(b[:self.byte_count])
            mask_to_opt = self._option_from_offset_mask.get(offset, {})
            m = 1
            for _ in range(8 * self.byte_count):
                if m in mask_to_opt:
                    values[str(int(mask_to_opt[m]))] = bool(value & m)
                m <<= 1
        return values

    def prepare_write(self, new_value):
        assert (isinstance(new_value, dict))
        w = {}
        for k, v in new_value.items():
            option = self._option_from_key[int(k)]
            if option.offset not in w:
                w[option.offset] = 0
            if v:
                w[option.offset] |= option.mask
        return [
            _int2bytes((offset << (8 * (2 * self.byte_count + 1)))
                       | (self.sep << (16 * self.byte_count))
                       | (self._mask_from_offset[offset] << (8 * self.byte_count))
                       | value, 2 * self.byte_count + 2) for offset, value in w.items()
        ]

    def all_options(self):
        return [int(opt) if isinstance(opt, int) else opt.as_int() for opt in self.options]


class ChoicesValidator(object):
    kind = KIND.choice
    """Translates between NamedInts and a byte sequence.
    :param choices: a list of NamedInts
    :param byte_count: the size of the derived byte sequence. If None, it
    will be calculated from the choices."""
    def __init__(self, choices, byte_count=None, read_skip_byte_count=None, write_prefix_bytes=b''):
        assert choices is not None
        assert isinstance(choices, _NamedInts)
        assert len(choices) > 1
        self.choices = choices
        self.needs_current_value = False

        max_bits = max(x.bit_length() for x in choices)
        self._byte_count = (max_bits // 8) + (1 if max_bits % 8 else 0)
        if byte_count:
            assert self._byte_count <= byte_count
            self._byte_count = byte_count
        assert self._byte_count < 8
        self._read_skip_byte_count = read_skip_byte_count if read_skip_byte_count else 0
        self._write_prefix_bytes = write_prefix_bytes if write_prefix_bytes else b''
        assert self._byte_count + self._read_skip_byte_count <= 14
        assert self._byte_count + len(self._write_prefix_bytes) <= 14

    def validate_read(self, reply_bytes):
        reply_value = _bytes2int(reply_bytes[self._read_skip_byte_count:self._read_skip_byte_count + self._byte_count])
        valid_value = self.choices[reply_value]
        assert valid_value is not None, '%s: failed to validate read value %02X' % (self.__class__.__name__, reply_value)
        return valid_value

    def prepare_write(self, new_value, current_value=None):
        if new_value is None:
            choice = self.choices[:][0]
        else:
            if isinstance(new_value, int):
                choice = self.choices[new_value]
            elif int(new_value) in self.choices:
                choice = self.choices[int(new_value)]
            elif new_value in self.choices:
                choice = self.choices[new_value]
            else:
                raise ValueError(new_value)

        if choice is None:
            raise ValueError('invalid choice %r' % new_value)
        assert isinstance(choice, _NamedInt)
        return self._write_prefix_bytes + choice.bytes(self._byte_count)


class ChoicesMapValidator(ChoicesValidator):
    kind = KIND.map_choice

    def __init__(
        self,
        choices_map,
        key_byte_count=None,
        byte_count=None,
        read_skip_byte_count=0,
        write_prefix_bytes=b'',
        extra_default=None,
        mask=-1,
        activate=0
    ):
        assert choices_map is not None
        assert isinstance(choices_map, dict)
        max_key_bits = 0
        max_value_bits = 0
        for key, choices in choices_map.items():
            assert isinstance(key, _NamedInt)
            assert isinstance(choices, list)
            max_key_bits = max(max_key_bits, key.bit_length())
            for key_value in choices:
                assert isinstance(key_value, _NamedInt)
                max_value_bits = max(max_value_bits, key_value.bit_length())
        self._key_byte_count = (max_key_bits + 7) // 8
        if key_byte_count:
            assert self._key_byte_count <= key_byte_count
            self._key_byte_count = key_byte_count
        self._byte_count = (max_value_bits + 7) // 8
        if byte_count:
            assert self._byte_count <= byte_count
            self._byte_count = byte_count
        self.choices = choices_map
        self.needs_current_value = False
        self.extra_default = extra_default
        self._read_skip_byte_count = read_skip_byte_count if read_skip_byte_count else 0
        self._write_prefix_bytes = write_prefix_bytes if write_prefix_bytes else b''
        self.activate = activate
        self.mask = mask
        assert self._byte_count + self._read_skip_byte_count + self._key_byte_count <= 14
        assert self._byte_count + len(self._write_prefix_bytes) + self._key_byte_count <= 14

    def validate_read(self, reply_bytes, key):
        start = self._key_byte_count + self._read_skip_byte_count
        end = start + self._byte_count
        reply_value = _bytes2int(reply_bytes[start:end]) & self.mask
        # reprogrammable keys starts out as 0, which is not a choice, so don't use assert here
        if self.extra_default is not None and self.extra_default == reply_value:
            return int(self.choices[key][0])
        assert reply_value in self.choices[
            key], '%s: failed to validate read value %02X' % (self.__class__.__name__, reply_value)
        return reply_value

    def prepare_write(self, key, new_value):
        choices = self.choices.get(key)
        if choices is None or (new_value not in choices and new_value != self.extra_default):
            _log.error('invalid choice %r for %s', new_value, key)
            return None
        new_value = new_value | self.activate
        return self._write_prefix_bytes + new_value.to_bytes(self._byte_count, 'big')


class RangeValidator(object):
    __slots__ = ('min_value', 'max_value', 'flag', '_byte_count', 'needs_current_value')

    kind = KIND.range
    """Translates between integers and a byte sequence.
    :param min_value: minimum accepted value (inclusive)
    :param max_value: maximum accepted value (inclusive)
    :param byte_count: the size of the derived byte sequence. If None, it
    will be calculated from the range."""
    def __init__(self, min_value, max_value, byte_count=None):
        assert max_value > min_value
        self.min_value = min_value
        self.max_value = max_value
        self.needs_current_value = False

        self._byte_count = math.ceil(math.log(max_value + 1, 256))
        if byte_count:
            assert self._byte_count <= byte_count
            self._byte_count = byte_count
        assert self._byte_count < 8

    def validate_read(self, reply_bytes):
        reply_value = _bytes2int(reply_bytes[:self._byte_count])
        assert reply_value >= self.min_value, '%s: failed to validate read value %02X' % (self.__class__.__name__, reply_value)
        assert reply_value <= self.max_value, '%s: failed to validate read value %02X' % (self.__class__.__name__, reply_value)
        return reply_value

    def prepare_write(self, new_value, current_value=None):
        if new_value < self.min_value or new_value > self.max_value:
            raise ValueError('invalid choice %r' % new_value)
        return _int2bytes(new_value, self._byte_count)


class MultipleRangeValidator:

    kind = KIND.multiple_range

    def __init__(self, items, sub_items):
        assert isinstance(items, list)  # each element must have .index and its __int__ must return its id (not its index)
        assert isinstance(sub_items, dict)
        # sub_items: items -> class with .minimum, .maximum, .length (in bytes), .id (a string) and .widget (e.g. 'Scale')
        self.items = items
        self._item_from_id = {int(k): k for k in items}
        self.sub_items = sub_items

    def prepare_read_item(self, item):
        return _int2bytes((self._item_from_id[int(item)].index << 1) | 0xFF, 2)

    def validate_read_item(self, reply_bytes, item):
        item = self._item_from_id[int(item)]
        start = 0
        value = {}
        for sub_item in self.sub_items[item]:
            r = reply_bytes[start:start + sub_item.length]
            if len(r) < sub_item.length:
                r += b'\x00' * (sub_item.length - len(value))
            v = _bytes2int(r)
            if not (sub_item.minimum < v < sub_item.maximum):
                _log.warn(
                    f'{self.__class__.__name__}: failed to validate read value for {item}.{sub_item}: ' +
                    f'{v} not in [{sub_item.minimum}..{sub_item.maximum}]'
                )
            value[str(sub_item)] = v
            start += sub_item.length
        return value

    def prepare_write(self, value):
        seq = []
        w = b''
        for item in value.keys():
            _item = self._item_from_id[int(item)]
            b = _int2bytes(_item.index, 1)
            for sub_item in self.sub_items[_item]:
                try:
                    v = value[str(int(item))][str(sub_item)]
                except KeyError:
                    return None
                if not (sub_item.minimum <= v <= sub_item.maximum):
                    raise ValueError(
                        f'invalid choice for {item}.{sub_item}: {v} not in [{sub_item.minimum}..{sub_item.maximum}]'
                    )
                b += _int2bytes(v, sub_item.length)
            if len(w) + len(b) > 15:
                seq.append(b + b'\xFF')
                w = b''
            w += b
        seq.append(w + b'\xFF')
        return seq

    def prepare_write_item(self, item, value):
        _item = self._item_from_id[int(item)]
        w = _int2bytes(_item.index, 1)
        for sub_item in self.sub_items[_item]:
            try:
                v = value[str(sub_item)]
            except KeyError:
                return None
            if not (sub_item.minimum <= v <= sub_item.maximum):
                raise ValueError(f'invalid choice for {item}.{sub_item}: {v} not in [{sub_item.minimum}..{sub_item.maximum}]')
            w += _int2bytes(v, sub_item.length)
        return w + b'\xFF'


# Turn diverted mouse movement events into a mouse gesture
#
# Uses the following FSM.
# At initialization, we go into `start` state and begin accumulating displacement.
# If terminated in this state, we report back no movement.
# If the mouse moves enough, we go into the `moved` state run the progress function.
# If terminated in this state, we report back how much movement.
class DivertedMouseMovement(object):
    def __init__(self, device, dpi_name, key):
        self.device = device
        self.key = key
        self.dx = 0.
        self.dy = 0.
        self.fsmState = 'idle'
        self.dpiSetting = next(filter(lambda s: s.name == dpi_name, device.settings), None)

    @staticmethod
    def notification_handler(device, n):
        """Called on notification events from the mouse."""
        if n.sub_id < 0x40 and device.features[n.sub_id] == _hidpp20.FEATURE.REPROG_CONTROLS_V4:
            state = device._divertedMMState
            assert state
            if n.address == 0x00:
                cid1, cid2, cid3, cid4 = _unpack('!HHHH', n.data[:8])
                state.handle_keys_event({cid1, cid2, cid3, cid4})
            elif n.address == 0x10:
                dx, dy = _unpack('!hh', n.data[:4])
                state.handle_move_event(dx, dy)

    def handle_move_event(self, dx, dy):
        # This multiplier yields a more-or-less DPI-independent dx of about 5/cm
        # The multiplier could be configurable to allow adjusting dx
        dpi = self.dpiSetting.read() if self.dpiSetting else 1000
        dx = float(dx) / float(dpi) * 15.
        self.dx += dx
        dy = float(dy) / float(dpi) * 15.
        self.dy += dy
        if self.fsmState == 'pressed':
            if abs(self.dx) >= 1. or abs(self.dy) >= 1.:
                self.fsmState = 'moved'

    def handle_keys_event(self, cids):
        if self.fsmState == 'idle':
            if self.key in cids:
                self.fsmState = 'pressed'
                self.dx = 0.
                self.dy = 0.
        elif self.fsmState == 'pressed' or self.fsmState == 'moved':
            if self.key not in cids:
                # emit mouse gesture notification
                from .base import _HIDPP_Notification as _HIDPP_Notification
                from .common import pack as _pack
                from .diversion import process_notification as _process_notification
                payload = _pack('!hh', int(self.dx), int(self.dy))
                notification = _HIDPP_Notification(0, 0, 0, 0, payload)
                _process_notification(self.device, self.device.status, notification, _hidpp20.FEATURE.MOUSE_GESTURE)
                self.fsmState = 'idle'
                self.dx = 0.
                self.dy = 0.


MouseGestureKeys = [
    _special_keys.CONTROL.Mouse_Gesture_Button,
    _special_keys.CONTROL.MultiPlatform_Gesture_Button,
]


class DivertedMouseMovementRW(object):
    def __init__(self, dpi_name, divert_name):
        self.kind = FeatureRW.kind  # pretend to be FeatureRW as required for HID++ 2.0 devices
        self.dpi_name = dpi_name
        self.divert_name = divert_name
        self.key = None

    def read(self, device):  # need to return bytes, as if read from device
        return _int2bytes(device._divertedMMState.key, 2) if '_divertedMMState' in device.__dict__ else b'\x00\x00'

    def write(self, device, data_bytes):
        def handler(device, n):
            """Called on notification events from the mouse."""
            if n.sub_id < 0x40 and device.features[n.sub_id] == _hidpp20.FEATURE.REPROG_CONTROLS_V4:
                state = device._divertedMMState
                if n.address == 0x00:
                    cid1, cid2, cid3, cid4 = _unpack('!HHHH', n.data[:8])
                    state.handle_keys_event({cid1, cid2, cid3, cid4})
                elif n.address == 0x10:
                    dx, dy = _unpack('!hh', n.data[:4])
                    state.handle_move_event(dx, dy)

        key = _bytes2int(data_bytes)
        if key:  # enable
            # Enable HID++ events on moving the mouse while button held
            self.key = next((k for k in device.keys if k.key == key), None)
            if self.key:
                self.key.set_rawXY_reporting(True)
                divertSetting = next(filter(lambda s: s.name == self.divert_name, device.settings), None)
                divertSetting.write_key_value(int(self.key.key), 1)
                from solaar.ui import status_changed as _status_changed
                _status_changed(device, refresh=True)  # update main window
                # Store our variables in the device object
                device._divertedMMState = DivertedMouseMovement(device, self.dpi_name, self.key.key)
                device.add_notification_handler('diverted-mouse-movement-handler', handler)
                return True
            else:
                _log.error('cannot enable diverted mouse movement on %s for key %s', device.name, key)
        else:  # disable
            if self.key:
                self.key.set_rawXY_reporting(False)
                divertSetting = next(filter(lambda s: s.name == self.divert_name, device.settings), None)
                divertSetting.write_key_value(int(self.key.key), 0)
                from solaar.ui import status_changed as _status_changed
                _status_changed(device, refresh=True)  # update main window
                self.key = None
            try:
                device.remove_notification_handler('diverted-mouse-movement-handler')
            except Exception:
                pass
            if hasattr(device, '_divertedMMState'):
                del device._divertedMMState
        return True


def apply_all_settings(device):
    persister = getattr(device, 'persister', None)
    sensitives = persister.get('_sensitive', {}) if persister else {}
    for s in device.settings:
        ignore = sensitives.get(s.name, False)
        if ignore != SENSITIVITY_IGNORE:
            s.apply()
